@@ -247,7 +247,8 @@ Value *SPIRVExpander::visitCallInst(CallInst &CI) {
   }
   // SPV_KHR_shader_clock extension.
   if (CalleeName.startswith("ReadClockKHR")) {
-    auto *Intr = emitIntrinsic(Builder, Intrinsic::readcyclecounter, llvm::ArrayRef<llvm::Type *>(), {});
+    auto *Intr = emitIntrinsic(Builder, Intrinsic::readcyclecounter,
+                               llvm::ArrayRef<llvm::Type *>(), {});
     return Builder.CreateBitCast(Intr, CI.getType());
   }
   // SPV_EXT_shader_atomic_float_min_max extension
