@@ -1,6 +1,6 @@
 /*========================== begin_copyright_notice ============================
 
-Copyright (C) 2017-2023 Intel Corporation
+Copyright (C) 2017-2025 Intel Corporation
 
 SPDX-License-Identifier: MIT
 
@@ -33,7 +33,8 @@ namespace llvm
     class Value;
     class GlobalVariable;
     class StructType;
-}
+    class Type;
+    }
 
 const unsigned int INPUT_RESOURCE_SLOT_COUNT = 128;
 const unsigned int NUM_SHADER_RESOURCE_VIEW_SIZE = (INPUT_RESOURCE_SLOT_COUNT + 1) / 64;
@@ -268,6 +269,10 @@ enum class ShaderTypeMD
         // existence of this value indicates we are opting into atomic pull tile walk
         // this field carries the uber tile dimensions, which needs to be passed on to the UMD.
         std::optional<std::array<uint32_t, 2>> uberTileDimensions;
+
+        // this fields marks how many hw stacks are needed to satisfy every shader in the module
+        // its std::max of numSyncRTStacks in all shaders
+        uint32_t numSyncRTStacks = 0;
     };
 
     // Info specific to each raytracing shader
@@ -313,6 +318,10 @@ enum class ShaderTypeMD
         // a collection of other names that they go by.
         std::vector<std::string> Aliases;
 
+        // this fields marks how many hw stacks the shader needs for its rayqueries
+        // some optimizations might want to know that
+        uint32_t numSyncRTStacks = 0;
+
         // for continuations used in ReorderThread, this field indicates the maximum value of the coherence hint
         uint32_t NumCoherenceHintBits = 0;
     };
@@ -352,6 +361,8 @@ enum class ShaderTypeMD
         bool globalIDPresent = false;
         // This is true if the function has any sync raytracing functionality
         bool hasSyncRTCalls = false;
+        bool hasPrintfCalls = false;
+        bool hasIndirectCalls = false;
 
         // Analysis result of if there are non-kernel-argument ld/st in the kernel
         bool hasNonKernelArgLoad = false;
@@ -404,8 +415,9 @@ enum class ShaderTypeMD
         int LoadCacheDefault                            = -1;
         int StoreCacheDefault                           = -1;
 
-        unsigned VISAPreSchedRPThreshold           = 0;
-        unsigned SetLoopUnrollThreshold            = 0;
+        unsigned VISAPreSchedRPThreshold                = 0;
+        unsigned VISAPreSchedCtrl                       = 0;
+        unsigned SetLoopUnrollThreshold                 = 0;
         bool UnsafeMathOptimizations                    = false;
         bool disableCustomUnsafeOpts                    = false;
         bool disableReducePow                           = false;
@@ -427,6 +439,7 @@ enum class ShaderTypeMD
         bool BufferOffsetArgOptional                    = true;
         bool replaceGlobalOffsetsByZero                 = false;
         unsigned forcePixelShaderSIMDMode               = 0;
+        unsigned int forceTotalGRFNum                   = 0; // 0 means not forced
         bool pixelShaderDoNotAbortOnSpill               = false;
         bool UniformWGS                                 = false;
         bool disableVertexComponentPacking              = false;
@@ -491,6 +504,11 @@ enum class ShaderTypeMD
         bool WaDisablePayloadCoalescing                 = false;
         unsigned Quad8InputThreshold                    = 0;
         bool UseResourceLoopUnrollNested                = false;
+        bool DisableLoopUnroll                          = false;
+        unsigned ForcePushConstantMode                  = 0;
+        bool UseInstructionHoistingOptimization         = false;
+        bool DisableResourceLoopDestLifeTimeStart       = false;
+        bool UseLinearScanRA                            = false;
     };
 
     enum class ThreadIDLayout
@@ -511,6 +529,7 @@ enum class ShaderTypeMD
         unsigned char forcedSIMDSize = 0;  // 0 means not forced
         unsigned int forceTotalGRFNum = 0; // 0 means not forced
         unsigned int VISAPreSchedRPThreshold = 0; // 0 means use the default
+        unsigned int VISAPreSchedCtrl = 0; // 0 means use the default
         unsigned int SetLoopUnrollThreshold = 0; // 0 means use the default
         bool forceSpillCompression = false;
         bool allowLowerSimd = false;
@@ -846,6 +865,7 @@ enum class ShaderTypeMD
     bool isContinuation(const IGC::FunctionMetaData& funcMD);
     bool isCallStackHandler(const IGC::FunctionMetaData &funcMD);
 
+
     // User annotations query functions
     int extractAnnotatedNumThreads(const IGC::FunctionMetaData& funcMD);
 }
