@@ -1234,6 +1234,7 @@ struct CustomHash {
   }
 };
 
+[[maybe_unused]]
 static bool operator==(const Interval &first, const Interval &second) {
   return first.start->getLexicalId() == second.start->getLexicalId();
 }
@@ -1311,7 +1312,6 @@ public:
     forbiddenVec[(size_t)forbiddenKind::FBD_SCALAR].resize(
         getForbiddenVectorSize(G4_SCALAR));
   };
-  ~ForbiddenRegs(){};
 
   unsigned getForbiddenVectorSize(G4_RegFileKind regKind) const;
   void generateReservedGRFForbidden(unsigned reserveSpillSize);
@@ -1954,15 +1954,18 @@ public:
   void removeUnreferencedDcls();
   LocalLiveRange *GetOrCreateLocalLiveRange(G4_Declare *topdcl);
 
-  GlobalRA(G4_Kernel &k, PhyRegPool &r, PointsToAnalysis &p2a)
-      : kernel(k), builder(*k.fg.builder), regPool(r), pointsToAnalysis(p2a),
-        fbdRegs(*k.fg.builder), incRA(*this),
-        useLscForSpillFill(k.fg.builder->supportsLSC()),
-        useLscForNonStackCallSpillFill(
-            k.fg.builder->useLscForNonStackSpillFill()),
-        useLscForScatterSpill(k.fg.builder->supportsLSC() &&
-                              k.fg.builder->getOption(vISA_scatterSpill)),
-        use4GRFAlign(k.fg.builder->supports4GRFAlign()) {
+  GlobalRA(G4_Kernel& k, PhyRegPool& r, PointsToAnalysis& p2a)
+    : fbdRegs(*k.fg.builder),
+      use4GRFAlign(k.fg.builder->supports4GRFAlign()),
+      kernel(k), builder(*k.fg.builder), regPool(r),
+      pointsToAnalysis(p2a),
+      useLscForSpillFill(k.fg.builder->supportsLSC()),
+      useLscForScatterSpill(k.fg.builder->supportsLSC() &&
+          k.fg.builder->getOption(vISA_scatterSpill)),
+      useLscForNonStackCallSpillFill(
+          k.fg.builder->useLscForNonStackSpillFill()),
+      incRA(*this)
+  {
     vars.resize(k.Declares.size());
 
     if (kernel.getOptions()->getOption(vISA_VerifyAugmentation)) {
@@ -2059,6 +2062,7 @@ public:
   void verifySpillFill();
   void resetGlobalRAStates();
   bool canSkipFDE() const;
+  void initSRAsScratch() const;
 
   void insertPhyRegDecls();
 
